PAGE	50,132
TITLE	KEYSLAVE Version 0.1, September 1991
COMMENT	\

Accepts bytes from the parallel port, assembles them into words and
stores them in the keyboard buffer.

To make the transfer faster, speed the clock up by 4, and call
the original 08 handler only every 4th tick.

Version 0.1, written for January 1991 SAR
(c) 1991, BYTE/McGraw Hill
Portions (c) 1987, 1990 Overpriced Software

Version 0.1	HME	Initial code

Assemble KEYSLAVE.ASM under Turbo Assembler with:
	tasm /ml keyslave
	tlink /x /c /t keyslave
If you're running with Turbo Debugger:
	tasm /ml /la /zi keyslave
	tlink /m /l /s /c /v /t keyslave
	tdmap keyslave.map keyslave.tds
\

CODE SEGMENT PUBLIC 'CODE'

	ORG	100H

ASSUME	CS:CODE,DS:CODE,ES:NOTHING,SS:NOTHING

KEYSLAVE	PROC	FAR

START:	JMP	SET_VECTOR

ID_string	db	'Keyboard emulator 0.1'	; Who we are - for LOOK4ME
ID_loaded	dw	0			; after loading,
						; we'll turn it to a '-1'
ID_string_len	equ	$-START
;
SPEEDUP	equ	1		; optionally speed up the clock
TSR	equ	1		; Load as TSR or run as COM
DEBUG	equ	0		; Turns on a HEX display in the
				; upper left
PS2	equ	0		; PS/2 and compatible BIOSes have a function
				; to store data in the keyboard
				; buffer.  Older machines don't.
COMFILE	equ	1-TSR

	if	DEBUG
SCRN_SEG	equ	0b000h	; B000 for mono, B800 for CGA->VGA
	endif
LPT1_addr	equ	0378h	; LPT1
LPT2_addr	equ	0278h	; LPT2
Mono_addr	equ	03BCh	; LPT1 on mono cards
; Laptop output bits mapped as follows
;	0	Nybble select 0
;	1	Nybble select 1
; Laptop input bits (address+1)
;	bit#	function			original function
;	3	low bit from nybble select	ERROR
;		(data return)
;	4	Nybble data 0			Select
;	5	Nybble data 1			Paper out
;	6	Nybble data 2			ACK
;	7	Nybble data 3 (invert)		BUSY
Data_mask	equ	0f0h	; isolate the incoming data
Data_adj	equ	080h	; XOR to adjust the data polarity
Nybble_ret	equ	008h	; isolate the nybble return
Keyb_status_msk	equ	10h
Keyb_status	db	0
switch_val	db	0	; 1=LPT1, 2=LPT2, m=LPT1 (mono)
Printer_addr	dw	LPT1_addr
hex_table	db	'0123456789ABCDEF'

Input_data	dw	0

; data in segment 40h - keyboard info
buffer_head	equ	01ah
buffer_tail	equ	01ch
buffer_start	equ	80h
buffer_end	equ	82h
state_curr	db	0
state_table	dw	ST_0,ST_1,ST_2,ST_3,ST_4

; The clock is running 72 times/second.  Divide by 4 to get the original
; 18 ticks per second, divide by 9 to set the keyboard status 8 times/second
; 

status_div	db	2	; get the keyboard status every 8
				; ticks
clock_div	db	0	; reset and call old 08 handler
				; every 4 ticks
Timer_ctrl	equ	43h
Timer_data	equ	40h
EOI		equ	20h
;---- Timer interrupt routine

ASSUME	CS:CODE,DS:CODE
;
int	PROC	FAR
	push	ax
	push	bx
	push	cx
	push	dx
	push	si
	push	ds
	push	es

	push	cs
	pop	ds

;
; Now, let's see what there is to do.
; Process the next tick, based on the previous state
	if	COMFILE
COM_LOOP:
	endif
	mov	ax,Input_data
	if	DEBUG
	call	disp_val
	endif
;
	mov	bl,state_curr
	mov	bh,0
	shl	bx,1
	add	bx,offset state_table
	jmp	[bx]
ST_0:
; see if nybble 0 is available
	mov	al,0
	or	al,Keyb_status
	call	PORTOUT
	call	PORTIN
; if the data return bit is the same as the low bit of the address
; (in this case, a 0), then the data is valid
	mov	ah,al		; make a copy
	and	al,Nybble_ret
	jz	ST_0a
	jmp	int_out		; not yet
ST_0a:
; The data is the high bits of AH, adjusted
	and	ah,Data_mask
	xor	ah,Data_adj
	shr	ah,1
	shr	ah,1
	shr	ah,1
	shr	ah,1
	mov	al,ah
	xor	ah,ah		; AX is most 0, but the low nybble
	mov	Input_data,ax
	inc	state_curr
	mov	al,1
	or	al,Keyb_status
	call	PORTOUT
	jmp	int_out
ST_1:
; see if nybble 1 is available
	mov	al,1
	or	al,Keyb_status
	call	PORTOUT
	call	PORTIN
; if the data return bit is the same as the low bit of the address
; (in this case, a 1), then the data is valid
	mov	ah,al		; make a copy
	and	al,Nybble_ret
	jnz	ST_1a
	jmp	int_out		; not yet
ST_1a:
; The data is the high bits of AH, adjusted
	and	ah,Data_mask
	xor	ah,Data_adj
	mov	bl,ah
	xor	bh,bh
	mov	ax,Input_data
	add	ax,bx
	mov	Input_data,ax
	inc	state_curr
	mov	al,2
	or	al,Keyb_status
	call	PORTOUT
	jmp	int_out
ST_2:
; see if nybble 2 is available
	mov	al,2
	or	al,Keyb_status
	call	PORTOUT
	call	PORTIN
; if the data return bit is the same as the low bit of the address
; (in this case, a 0), then the data is valid
	mov	ah,al		; make a copy
	and	al,Nybble_ret
	jnz	int_out		; not yet
; The data is the high bits of AH, adjusted
	and	ah,Data_mask
	xor	ah,Data_adj
	shr	ah,1
	shr	ah,1
	shr	ah,1
	shr	ah,1
	mov	bh,ah
	xor	bl,bl
	mov	ax,Input_data
	add	ax,bx
	mov	Input_data,ax
	inc	state_curr
	mov	al,3
	or	al,Keyb_status
	call	PORTOUT
	jmp	int_out
ST_3:
; see if nybble 3 is available
	mov	al,3
	or	al,Keyb_status
	call	PORTOUT
	call	PORTIN
; if the data return bit is the same as the low bit of the address
; (in this case, a 0), then the data is valid
	mov	ah,al		; make a copy
	and	al,Nybble_ret
	jz	int_out		; not yet
; The data is the high bits of AH, adjusted
	and	ah,Data_mask
	xor	ah,Data_adj
	mov	bh,ah
	xor	bl,bl
	mov	ax,Input_data
	add	ax,bx
	mov	Input_data,ax
	mov	state_curr,4		; in case the buffer's full
ST_4:
; output it
	mov	dx,ax
	if	PS2
; PS/2s have INT 16, Func 5
	mov	ah,5
	int	16h
	else
; XTs and older ATs don't
	call	store_buffer
	endif
; If al=0, no problem.  If al=1 then buffer was full
	cmp	al,0
	jne	int_out		; couldn't save it
	mov	state_curr,al	; back to state 0
	or	al,Keyb_status
	call	PORTOUT
int_out:
; Now check the clocks
	inc	status_div
	mov	al,status_div
	if	SPEEDUP
	cmp	al,8
	else
	cmp	al,2
	endif
	jne	chk_timer
	mov	status_div,0
	mov	ah,1
	int	16h
; The zero flag is set if the buffer is empty. Get it into AH
	lahf
	xor	ah,40h		; toggle it
	shr	ah,1		; move it from bit 6 to bit 4
	shr	ah,1
	and	ah,Keyb_status_msk
	mov	Keyb_status,ah
chk_timer:
	if COMFILE
	jmp	COM_LOOP
	endif
	if	SPEEDUP
	inc	clock_div
	mov	al,clock_div
	cmp	al,4
	je	int_out2
; don't call the old one. It's not the 4th tick
	mov	al,EOI
	out	020h,al

	pop	es
	pop	ds
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	iret
	endif
;

int_out2:
; This is the 4th tick.  Call the old one.
	mov	clock_div,0

	pop	es
	pop	ds
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	DB	0EAH		; THIS INFORMATION IS FILLED IN AT SETUP TIME
top08	DW	00		; TOPS HOLDS THE IP REGISTER
bot08	DW	00		; BOTTOM HOLDS THE CS REGISTER
;
int	ENDP

int17	PROC	FAR
; Eat the printer interrupts and report everything OK.

; on entry, AH=function.
; AH=0	Print character
; AH=1	Init printer
; AH=2	Read status
	cmp	ah,0
	je	f17_0
	cmp	ah,1
	je	f17_1
	cmp	ah,2
	je	f17_1
	jmp	int17_out
f17_0:
f17_1:
f17_2:
; Fake out functions 0-2 to look like it worked
	mov	ah,90h		; report everything OK & not busy
int17_out:
	iret
;
int17	ENDP
;
;-----------------------------------------------------------------
store_buffer	proc	near
; Scan code/ascii char in DX is to be stored in the keyboard buffer
; Returns: AL=0 if OK, AL=1 if buffer full
	push	bx
	push	si
	push	di
	push	es

	cli			; Please don't interrupt
	mov	ax,40h		; keyboard data in segment 40
	mov	es,ax
	mov	di,es:[buffer_tail]
	mov	bx,di		; this is where we'll store to
	add	di,2		; point to next location
	mov	ax,es:[buffer_end]
; Some machines don't store the buffer start/end in [80],[82h]
	cmp	ax,0
	jne	sb_0
; For example, this one!!
	mov	ax,3eh		; the 'standard' buffer_end
sb_0:
	cmp	di,ax
	jne	sb_nowrap
	mov	di,es:[buffer_start]
	cmp	di,0		; legit _start?
	jne	sb_nowrap	; looks like it
	mov	di,1eh		; the 'standard buffer_start
sb_nowrap:
	mov	ax,es:[buffer_head]
	cmp	ax,di		; If =, then no more room
	je	sb_err
	mov	es:[bx],dx
	mov	es:[buffer_tail],di
	mov	al,0
	jmp	short sb_out
sb_err:
	mov	al,1
sb_out:
	sti			; interrupts back on
	pop	es
	pop	di
	pop	si
	pop	bx
	ret
store_buffer	endp
;-----------------------------------------------------------------
	if	DEBUG
disp_val	proc	near
; Convert to Ascii
	push	ax
	push	cx
	push	bx
	push	es

; High byte first. Save the low byte
	mov	cl,al
	mov	al,ah
	and	al,0f0h
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	mov	bx,offset hex_table
	xlat	hex_table
	xchg	al,ah
	and	al,0fh
	xlat	hex_table
; display it
	push	ax		; high nybble in AH, low in AL
	push	ax
	mov	ax,SCRN_SEG
	mov	es,ax
	pop	ax
	mov	al,ah
	mov	ah,7
	mov	word ptr es:[0],ax
	pop	ax
	mov	ah,7
	mov	word ptr es:[2],ax
; No the low byte
	mov	al,cl
	mov	ah,al
	and	al,0f0h
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	mov	bx,offset hex_table
	xlat	hex_table
	xchg	al,ah
	and	al,0fh
	xlat	hex_table
; display it
	push	ax		; high nybble in AH, low in AL
	push	ax
	mov	ax,SCRN_SEG
	mov	es,ax
	pop	ax
	mov	al,ah
	mov	ah,7
	mov	word ptr es:[4],ax
	pop	ax
	mov	ah,7
	mov	word ptr es:[6],ax

	pop	es
	pop	bx
	pop	cx
	pop	ax
	ret
disp_val	endp
	endif
;-----------------------------------------------------------------
PORTIN	proc	near
; Grabs a byte from the printer port
; OUTPUT:
;	AL=data byte
	mov	ax,Printer_addr
	inc	ax
	mov	dx,ax
	in	al,dx
	ret
PORTIN	endp
;-----------------------------------------------------------------
PORTOUT	proc	near
; Output a byte to the printer port
; INPUT:
;	AL=data byte
	mov	dl,al
	mov	ax,Printer_addr
	xchg	ax,dx
	out	dx,al
	ret
PORTOUT	endp
;-----------------------------------------------------------------
int_end		LABEL	WORD
;
SET_VECTOR:
;
; Before we start, look through the command line for a '/'.  The next
; character is the LPT port
	mov	bx,81h
cl_0:
	mov	al,[bx]
	cmp	al,13		; CR?
	je	bad_switch
	cmp	al,'/'
	je	test_char
	inc	bx		; not yet
	jmp	short cl_0
test_char:
	mov	al,[bx+1]
	cmp	al,'m'
	je	is_mono
	cmp	al,'1'
	jl	bad_switch
	cmp	al,'2'
	jg	bad_switch
	mov	switch_val,al
	mov	msg_1a,al
	mov	msg_3a,al		; update the two messages
	sub	al,'1'			; Now, 0 or 1
; LPT1 is 378, LPT2 is 278.  Multiply AL by 100H and subtract
; from 378
	mov	cl,8
	cbw
	shl	ax,cl
	neg	ax
	add	ax,378h
	mov	Printer_addr,ax
	jmp	short seg_scan
is_mono:
	mov	ax,Mono_addr
	mov	Printer_addr,ax
	jmp	short seg_scan
bad_switch:
	mov	dx,offset msg_4
	mov	ah,9
	int	21h
	jmp	bye_bye

seg_scan:	
	if	TSR
; Start scanning through each segment, comparing ID_string_len
; bytes to our current segment.  If we find a match, AND it's not CS,
; we're already loaded
;
	mov	ax,cs
	mov	dx,ax			; keep a copy of CS around.
; Before we start, change ID_loaded to 1.  The copy in the DOS disk
; buffer will contain ID_loaded=0.  We don't want to match it.
	mov	cs:ID_loaded,-1
	mov	ax,1			; first segment to test
Look4Me:
	mov	es,ax
	call	chk_ID
	jz	already_loaded		; found it. Darn.
	inc	ax
	cmp	ax,dx
	jne	Look4Me			; As long as we're not
					; testing the current CS, keep looking.
	endif
Start_load:
	if	COMFILE
	jmp	COM_LOOP
	endif
; Do the install.
;
	push	cs
	pop	ds

	mov	ax,3508h		; set up vector table
	int	21h			; use dos to get vector
	push	es
	pop	ax
	push	cs
	pop	ds
	mov	bot08,ax
	push	bx
	pop	ax
	mov	top08,ax

	mov	ax,2508h		; set up vector table
	mov	bx,cs
	mov	ds,bx
	mov	dx,offset int		; set it to our interrupt
	int	21h			; use dos to set vector

	mov	ax,2517h		; set up fake printer BIOS
	mov	bx,cs
	mov	ds,bx
	mov	dx,offset int17		; set it to our interrupt
	int	21h			; use dos to set vector

	if	SPEEDUP
; Speed up the clock by 4.  The old divisor was 65535, We want 16384.
	cli
	mov	al,36h			; mode 3, LSB then MSB
	out	Timer_ctrl,al
	jmp	short $+2		; brief I/O delay
	jmp	short $+2		; brief I/O delay
	mov	al,00h			; LSB of 16384
	out	Timer_data,al
	jmp	short $+2		; brief I/O delay
	jmp	short $+2
	mov	al,40h			; MSB of 16384
	out	Timer_data,al
	sti
	endif

; put up a little message.
	mov	dx,offset msg_1
	mov	bx,cs
	mov	ds,bx
	mov	ah,9
	int	21h			; display the message
; terminate but stay resident

	mov	bx,cs
	mov	ds,bx
	mov	dx,offset int_end+1	; point to the end of the program
	int	27h

already_loaded:
	if	SPEEDUP
; Speed up the clock by 4.  The old divisor was 65535, We want 16384.
	cli
	mov	al,36h			; mode 3, LSB then MSB
	out	Timer_ctrl,al
	jmp	short $+2		; brief I/O delay
	jmp	short $+2		; brief I/O delay
	mov	al,00h			; LSB of 16384
	out	Timer_data,al
	jmp	short $+2		; brief I/O delay
	jmp	short $+2
	mov	al,40h			; MSB of 16384
	out	Timer_data,al
	sti
	endif
; put up a little message.
	mov	dx,offset msg_3
al_1:
	mov	bx,cs
	mov	ds,bx
	mov	ah,9
	int	21h			; display the message
; ES points to where PG loaded before.  Change the port assignment
	mov	ax,Printer_addr
	mov	es:Printer_addr,ax
bye_bye:
	int	20h
;-----------------------------------------------------------------
; Installation subroutines
;-----------------------------------------------------------------
; Compare the string in CS:START to ES:START.  If it matches all ID_string_len
; bytes, return with Z=1, else Z=0
chk_ID	proc	near
		push	cs
		pop	ds			; It should already be
						; but make sure
;
		mov	si,offset START		
		mov	di,si			;  fingerprint.
		mov	cx,ID_string_len
		repe	cmpsb
		jnz	ci_0
		nop
ci_0:
		ret
chk_ID	endp
;-----------------------------------------------------------------
KEYSLAVE		ENDP
msg_1	db	'KEYSLAVE V0.1 is now installed on LPT'
msg_1a	db	'm:.',13,10
	db	'(c) 1991, Overpriced Software & BYTE/McGraw Hill',13,10
	db	'Portions (c) 1987, 1990 Overpriced Software',13,10,'$'
msg_3	db	'Switching to LPT'
msg_3a	db	'm:',13,10,'$'
msg_4	db	'Usage: KEYSLAVE /port, where port=1 for LPT1:',13,10
	db	'                                2 for LPT2:',13,10
	db	'                                m for Monochrome adapter port'
	db	13,10,'$'
;
;
PROG_END	LABEL	BYTE
CODE		ENDS
	END	START		; JUMP TO START OF THE PROGRAM
